/*
 *  wavsplit by Davide Libenzi ( WAV audio file splitter )
 *  Copyright (C) 2003  Davide Libenzi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#if !defined(__GNUC__)
#define strcasecmp stricmp
#endif



typedef signed char wav_int8;
typedef unsigned char wav_uint8;
typedef signed short wav_int16;
typedef unsigned short wav_uint16;
typedef signed int wav_int32;
typedef unsigned int wav_uint32;

#if defined(__GNUC__)
typedef signed long long wav_int64;
typedef unsigned long long wav_uint64;
#else
typedef signed __int64 wav_int64;
typedef unsigned __int64 wav_uint64;
#endif



typedef struct s_wavefile {
	char *fname;
	FILE *file;
	wav_uint16 aufmt;
	wav_uint16 numch;
	wav_uint16 blkalign;
	wav_uint16 smpbit;
	wav_uint32 c1siz;
	wav_uint32 sbc1siz;
	wav_uint32 smprate;
	wav_uint32 bytrate;
	wav_uint32 sbc2siz;
	wav_uint32 sbc2rd;
	wav_uint32 wrsiz;
} wavefile_t;



static int wav_read16(FILE *file, wav_uint16 *val) {
	wav_uint8 buf[2];

	if (!fread(buf, 2, 1, file))
		return -1;
	*val = ((wav_uint16) buf[0]) | ((wav_uint16) buf[1]) << 8;
	return 0;
}


static int wav_read32(FILE *file, wav_uint32 *val) {
	wav_uint8 buf[4];

	if (!fread(buf, 4, 1, file))
		return -1;
	*val = ((wav_uint32) buf[0]) | ((wav_uint32) buf[1]) << 8 |
		((wav_uint32) buf[2]) << 16 | ((wav_uint32) buf[3]) << 24;
	return 0;
}


static int wav_write16(FILE *file, wav_uint16 val) {
	wav_uint8 buf[2];

	buf[0] = (wav_uint8) val;
	buf[1] = (wav_uint8) (val >> 8);
	if (!fwrite(buf, 2, 1, file))
		return -1;
	return 0;
}


static int wav_write32(FILE *file, wav_uint32 val) {
	wav_uint8 buf[4];

	buf[0] = (wav_uint8) val;
	buf[1] = (wav_uint8) (val >> 8);
	buf[2] = (wav_uint8) (val >> 16);
	buf[3] = (wav_uint8) (val >> 24);
	if (!fwrite(buf, 4, 1, file))
		return -1;
	return 0;
}


static int wav_open(wavefile_t *wf, char const *fname) {
	FILE *file;
	wav_uint8 buf[4];

	if (!(file = fopen(fname, "rb"))) {
		perror(fname);
		return -1;
	}
	if (!fread(buf, 4, 1, file) || memcmp(buf, "RIFF", 4) ||
	    wav_read32(file, &wf->c1siz) < 0 || !fread(buf, 4, 1, file) ||
	    memcmp(buf, "WAVE", 4) || !fread(buf, 4, 1, file) || memcmp(buf, "fmt ", 4) ||
	    wav_read32(file, &wf->sbc1siz) < 0 || wav_read16(file, &wf->aufmt) < 0 ||
	    wav_read16(file, &wf->numch) < 0 || wav_read32(file, &wf->smprate) < 0 ||
	    wav_read32(file, &wf->bytrate) < 0 || wav_read16(file, &wf->blkalign) < 0 ||
	    wav_read16(file, &wf->smpbit) < 0) {
		fprintf(stderr, "invalid wav file: file=%s\n", fname);
		fclose(file);
		return -1;
	}

	if (fseek(file, 12 + (8 + wf->sbc1siz), SEEK_SET) || !fread(buf, 4, 1, file) ||
	    memcmp(buf, "data", 4) || wav_read32(file, &wf->sbc2siz) < 0) {
		fprintf(stderr, "invalid wav file: file=%s\n", fname);
		fclose(file);
		return -1;
	}
	wf->sbc2rd = 0;
	wf->fname = strdup(fname);
	wf->file = file;
	wf->wrsiz = 0;

	return 0;
}


static int wav_create(wavefile_t *wf, char const *fname, wav_uint16 aufmt, wav_uint16 numch,
		      wav_uint32 smprate, wav_uint32 bytrate, wav_uint16 blkalign, wav_uint16 smpbit) {
	FILE *file;

	if (!(file = fopen(fname, "w+b"))) {
		perror(fname);
		return -1;
	}
	if (!fwrite("RIFF", 4, 1, file) || wav_write32(file, 0) < 0 ||
	    !fwrite("WAVE", 4, 1, file) || !fwrite("fmt ", 4, 1, file) ||
	    wav_write32(file, 16) < 0 || wav_write16(file, aufmt) < 0 ||
	    wav_write16(file, numch) < 0 || wav_write32(file, smprate) < 0 ||
	    wav_write32(file, bytrate) < 0 || wav_write16(file, blkalign) < 0 ||
	    wav_write16(file, smpbit) < 0 || !fwrite("data", 4, 1, file) ||
	    wav_write32(file, 0) < 0) {
		perror(fname);
		fclose(file);
		return -1;
	}
	wf->aufmt = aufmt;
	wf->numch = numch;
	wf->smprate = smprate;
	wf->bytrate = bytrate;
	wf->blkalign = blkalign;
	wf->smpbit = smpbit;
	wf->sbc2siz = 0;
	wf->sbc2rd = 0;
	wf->fname = strdup(fname);
	wf->file = file;
	wf->wrsiz = 0;

	return 0;
}


static void wav_close(wavefile_t *wf) {

	if (wf->wrsiz) {
		if (fseek(wf->file, 4, SEEK_SET) ||
		    wav_write32(wf->file, 12 + (8 + 16) + (8 + wf->wrsiz)) < 0 ||
		    fseek(wf->file, 40, SEEK_SET) || wav_write32(wf->file, wf->wrsiz) < 0) {
			perror(wf->fname);
		}
	}
	fclose(wf->file);
	free(wf->fname);
}


static int wav_read_blks(wavefile_t *wf, wav_uint8 *data, int blks) {
	int rblk;
	wav_uint32 dsize, dread;
	wav_uint8 buf[4];

	dsize = (wav_uint32) blks * wf->blkalign;
	if (wf->sbc2siz) {
		if (wf->sbc2rd >= wf->sbc2siz) {
			if (!fread(buf, 4, 1, wf->file))
				return 0;
			if (memcmp(buf, "data", 4) || wav_read32(wf->file, &wf->sbc2siz) < 0) {
				fprintf(stderr, "invalid wav file: file=%s\n", wf->fname);
				return -1;
			}
			wf->sbc2rd = 0;
		}
		if (dsize + wf->sbc2rd > wf->sbc2siz)
			dsize = wf->sbc2siz - wf->sbc2rd;
	}
	dread = fread(data, 1, dsize, wf->file);
	if ((dread % wf->blkalign) != 0 || (wf->sbc2siz && dread != dsize)) {
		fprintf(stderr, "invalid wav file: file=%s\n", wf->fname);
		return -1;
	}
	rblk = (int) (dread / wf->blkalign);
	wf->sbc2rd += (wav_uint32) rblk * wf->blkalign;
	return rblk;
}


static int wav_write_blks(wavefile_t *wf, wav_uint8 const *data, int blks) {
	wav_uint32 dsize;

	dsize = (wav_uint32) blks * wf->blkalign;
	if (!fwrite(data, dsize, 1, wf->file)) {
		perror(wf->fname);
		return -1;
	}
	wf->wrsiz += dsize;
	return (int) (dsize / wf->blkalign);
}


static int wav_copy(wavefile_t *wfin, char const *fname, int blkcpy) {
	int blktx, bbufsiz, bcur, brd;
	wavefile_t wfout;
	wav_uint8 blkbuf[8192];

	if (wav_create(&wfout, fname, wfin->aufmt, wfin->numch,
		       wfin->smprate, wfin->bytrate, wfin->blkalign, wfin->smpbit) < 0)
		return -1;
	bbufsiz = sizeof(blkbuf) / wfin->blkalign;
	for (blktx = 0; blktx < blkcpy;) {
		bcur = blkcpy - blktx > bbufsiz ? bbufsiz: blkcpy - blktx;
		if ((brd = wav_read_blks(wfin, blkbuf, bcur)) < 0) {
			wav_close(&wfout);
			return -1;
		}
		if (brd == 0)
			break;
		if (wav_write_blks(&wfout, blkbuf, brd) != brd) {
			wav_close(&wfout);
			return -1;
		}
		blktx += brd;
	}
	wav_close(&wfout);

	return blktx;
}





static void usage(char const *name) {

	fprintf(stderr, "use:  %s --split-time nsec infile\n", name);
}


int main(int argc, char *argv[]) {
	int i, fcnt, blkfile, blkcpy;
	unsigned long spltime = 0;
	wav_uint64 bytfile;
	char *tmp;
	wavefile_t wfin;
	char bfname[320], cfname[320];

	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "--split-time")) {
			if (++i < argc)
				spltime = atol(argv[i]);
		} else
			break;
	}
	if (i == argc || spltime == 0) {
		usage(argv[0]);
		return 1;
	}

	if (wav_open(&wfin, argv[i]) < 0)
		return 2;

	bytfile = (wav_uint64) wfin.bytrate * (wav_uint64) spltime;
	blkfile = (int) (bytfile / wfin.blkalign);

	strncpy(bfname, argv[i], sizeof(bfname));
	if ((tmp = strrchr(bfname, '.')) != NULL && !strcasecmp(tmp, ".wav"))
		*tmp = '\0';

	for (fcnt = 1;; fcnt++) {
		sprintf(cfname, "%s-%d.wav", bfname, fcnt);
		if ((blkcpy = wav_copy(&wfin, cfname, blkfile)) != blkfile)
			break;
	}
	wav_close(&wfin);

	return 0;
}

